# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from datetime import datetime
from typing import Optional, List

from fhirtordf.rdfsupport.fhirgraphutils import value, extension, concept_uri, codeable_concept_code
from fhirtordf.rdfsupport.namespaces import FHIR, SNOMEDCT, V3
from fhirtordf.rdfsupport.uriutils import parse_fhir_resource_uri
from rdflib import Graph, URIRef, Literal, XSD

from i2fhirb2.fhir.fhirpatientmapping import FHIRPatientMapping
from i2fhirb2.i2b2model.data.i2b2codes import I2B2DemographicsCodes
from i2fhirb2.i2b2model.data.i2b2observationfact import ObservationFact, ObservationFactKey
from i2fhirb2.i2b2model.data.i2b2patientdimension import PatientDimension, VitalStatusCd
# TODO: is there any reason to pull patient_id from Patient.identifier rather than URL?
# TODO: what of foreign addresses?
from i2fhirb2.i2b2model.data.i2b2patientmapping import PatientIDEStatus
from i2fhirb2.sqlsupport.i2b2tables import I2B2Tables


class FHIRPatientDimension:
    def __init__(self, g: Graph, tables: Optional[I2B2Tables], patient: URIRef) -> None:
        """
        Generate i2b2 patient dimension and patient_mapping records from PATIENT resources in graph g

        :param g: Graph containing 0 or more FHIR Patient resources
        :param tables: i2b2 tables connection if we are using a database (vs. tsv output)
        :param patient: Graph subject
        """
        assert value(g, patient, FHIR.animal) is None       # We don't do animals
        parsed_resource = parse_fhir_resource_uri(patient)
        self.patient_mappings = FHIRPatientMapping(tables, parsed_resource.resource, str(parsed_resource.namespace))
        active = value(g, patient, FHIR.Patient.active)
        if active is not None and not active:
            self.patient_mappings._patient_ide_status = PatientIDEStatus.inactive
        self.patient_dimension_entry = PatientDimension(self.patient_mappings.patient_num,
                                                        VitalStatusCd(VitalStatusCd.bd_unknown,
                                                                      VitalStatusCd.dd_unknown))

        # Additional attributes that don't go into patient dimension but are recorded as observation facts
        self._language = None
        self._marital_status = None
        self._race = None
        self._religion = None
        self.add_patient_information(g, patient)

    def add_patient_information(self, g: Graph, patient: URIRef) -> None:
        """
        Add additional information to the patient
        :param g: Graph carrying additional facts about the patient
        :param patient: URI of the actual patient
        """
        if not g.value(patient, FHIR.Patient.animal):       # i2b2 doesn't do animals
            # gender
            gender = value(g, patient, FHIR.Patient.gender)
            if gender == "male":
                self.patient_dimension_entry._sex_cd = 'M'
            elif gender == "female":
                self.patient_dimension_entry._sex_cd = 'F'
            elif gender == "other":
                self.patient_dimension_entry._sex_cd = 'U'

            # deceased.deceasedBoolean --> vital_status_code.deathInd
            isdeceased = value(g, patient, FHIR.Patient.deceasedBoolean)
            if isdeceased is not None:
                self.patient_dimension_entry._vital_status_cd = VitalStatusCd.dd_deceased if isdeceased \
                    else VitalStatusCd.dd_living

            # deceased.deceasedDateTime --> deathcode / death_date
            self.deathdate = value(g, patient, FHIR.Patient.deceasedDateTime, True)

            # birthdate - must be processed after deceased, as deceased goes into age calculation
            bd = g.value(patient, FHIR.Patient.birthDate)
            birthdate = None
            if bd:
                birthdate = extension(g, bd, "http://hl7.org/fhir/StructureDefinition/patient-birthTime",
                                      asLiteral=True)
                if not birthdate:
                    birthdate = value(g, patient, FHIR.Patient.birthDate, asLiteral=True)
            self.birthdate = birthdate

            # address -- use == home / period.end is empty or past deathcode date
            addresses = g.objects(patient, FHIR.Patient.address)
            for address in addresses:
                address_use = value(g, address, FHIR.Address.use)
                if address_use is None or address_use == "home":
                    period = g.value(address, FHIR.Address.period)
                    if not period or (period and value(g, period, FHIR.Period.end) is None):
                        city = value(g, address, FHIR.Address.city)
                        state = value(g, address, FHIR.Address.state)
                        zipcode = value(g, address, FHIR.Address.postalCode)
                        if zipcode:
                            self.patient_dimension_entry._zip_cd = zipcode
                            if city and state:
                                self.patient_dimension_entry._statecityzip_path = \
                                    'Zip codes\\' + state + '\\' + city + '\\' + zipcode + '\\'

            # maritalStatus --> map to 'single', 'married', 'divorced', 'widow', other?
            marital_stati = codeable_concept_code(g, patient, FHIR.Patient.maritalStatus)
            for ms in marital_stati:
                if ms.system == str(V3.MaritalStatus):
                    self._marital_status = marital_stati[0]
                    msc = self._marital_status.code
                    if msc != 'UNK':
                        self.patient_dimension_entry._marital_status_cd = \
                            'divorced' if msc in ['A', 'D'] else \
                            'married' if msc in ['L', 'M', 'P'] else \
                            'widow' if msc in ['W'] else \
                            'single'
                    break
            else:
                msuri = concept_uri(g, patient, FHIR.Patient.maritalStatus, SNOMEDCT)
                if msuri:
                    # TODO: figure out what to do with SNOMED id's (terminology service, anyone?)
                    pass

            # language
            communications = list(g.objects(patient, FHIR.Patient.communication))
            language = None
            for communication in communications:
                pref = value(g, communication, FHIR.Patient.communication.preferred)
                if pref or (pref is None and len(communications) == 1):
                    languages = codeable_concept_code(g, communication, FHIR.Patient.communication.language)
                    if languages:
                        language = languages[0]
                        break

            if language is not None:
                self._language = language.code

            # race - not a part of the core fhir spec or known extensions

            # religion

    @property
    def birthdate(self) -> Literal:
        return Literal(self.patient_dimension_entry.birth_date)

    @birthdate.setter
    def birthdate(self, bd: Optional[Literal]) -> None:
        if bd is None:
            self.patient_dimension_entry._birth_date = None
        else:
            # TODO: decide whether we want to refine this further
            if bd.datatype == XSD.gYear:
                self.patient_dimension_entry._vital_status_code.birthcode = VitalStatusCd.bd_year
            elif bd.datatype == XSD.gYearMonth:
                self.patient_dimension_entry._vital_status_code.birthcode = VitalStatusCd.bd_month
            elif bd.datatype == XSD.date:
                self.patient_dimension_entry._vital_status_code.birthcode = VitalStatusCd.bd_day
            else:
                self.patient_dimension_entry._vital_status_code.birthcode = VitalStatusCd.bd_hour
            self.patient_dimension_entry._birth_date = bd.toPython()

            # Age calculation --
            # TODO: figure out what to do with tzoffset
            # TODO: figure out what rdflib delivers in the toPython() function
            niave_bd = bd.toPython()
            ref_date = None
            if self.patient_dimension_entry._vital_status_code.deathcode \
                    in {VitalStatusCd.dd_living, VitalStatusCd.dd_unknown}:
                ref_date = datetime.now()
            elif self.patient_dimension_entry._death_date is not None:
                ref_date = self.patient_dimension_entry._death_date
            if ref_date is not None:
                age = ref_date.year - niave_bd.year
                if age > 0:
                    if ref_date.month > niave_bd.month:
                        age -= 1
                    elif ref_date.month == niave_bd.month:
                        if ref_date.day > niave_bd.day:
                            age -= 1
                self.patient_dimension_entry._age_in_years_num = age

    @property
    def deathdate(self) -> Literal:
        return Literal(self.patient_dimension_entry.death_date)

    @deathdate.setter
    def deathdate(self, dd: Optional[Literal]) -> None:
        if dd is None:
            self.patient_dimension_entry._death_date = None
        else:
            # TODO: decide whether we want to refine this further
            if dd.datatype == XSD.gYear:
                self.patient_dimension_entry._vital_status_code.deathcode = VitalStatusCd.dd_year
            elif dd.datatype == XSD.gYearMonth:
                self.patient_dimension_entry._vital_status_code.deathcode = VitalStatusCd.dd_month
            elif dd.datatype == XSD.date:
                self.patient_dimension_entry._vital_status_code.deathcode = VitalStatusCd.dd_day
            else:
                self.patient_dimension_entry._vital_status_code.deathcode = VitalStatusCd.dd_hour
            self.patient_dimension_entry._death_date = dd.toPython()

    def as_observation_facts(self, encounter_num: int, provider_id: str, start_date: datetime) -> List[ObservationFact]:
        rval = []
        pde = self.patient_dimension_entry
        ofk = ObservationFactKey(self.patient_dimension_entry.patient_num, encounter_num, provider_id, start_date)

        # Age entry
        rval.append(ObservationFact(ofk, I2B2DemographicsCodes.age(pde.age_in_years_num)))

        # Sex
        rval.append(ObservationFact(ofk, 
                                    I2B2DemographicsCodes.sex_female if pde.sex_cd == 'F'
                                    else I2B2DemographicsCodes.sex_male if pde.sex_cd == 'M'
                                    else I2B2DemographicsCodes.sex_undifferentiated if pde.sex_cd == 'U'
                                    else I2B2DemographicsCodes.sex_unknown))

        # Birthdate
        if pde.birth_date:
            bd_of = ObservationFact(ofk, I2B2DemographicsCodes.birthdate)
            bd_of._date_val(pde.birth_date)
            rval.append(bd_of)

        # Deathdate
        if pde.death_date:
            dd_of = ObservationFact(ofk, I2B2DemographicsCodes.birthdate)
            dd_of._date_val(pde.death_date)
            rval.append(dd_of)

        # Language
        rval.append(ObservationFact(ofk, I2B2DemographicsCodes.language(self._language)))

        # Marital status
        rval.append(ObservationFact(ofk, I2B2DemographicsCodes.marital_status(self._marital_status)))

        # race
        rval.append(ObservationFact(ofk, I2B2DemographicsCodes.race(self._race)))

        # religion
        # Religion codes currently like the FHIR
        rval.append(ObservationFact(ofk, I2B2DemographicsCodes.religion(self._religion)))

        # vital -- no idea what vital_deferred means
        rval.append(
            ObservationFact(ofk,
                            I2B2DemographicsCodes.vital_living
                            if pde._vital_status_code.dd_deceased == VitalStatusCd.dd_living
                            else I2B2DemographicsCodes.vital_unknown if pde._vital_status_code.dd_unknown
                            else I2B2DemographicsCodes.vital_dead))

        # zip
        rval.append(ObservationFact(ofk, I2B2DemographicsCodes.zip(pde.zip_cd)))

        return rval



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import List, Optional, Union, Tuple

from fhirtordf.rdfsupport.namespaces import FHIR
from rdflib import URIRef, Graph, BNode, RDF
from rdflib.term import Node

from i2fhirb2.fhir.fhircodemapping import process_concept_code
from i2fhirb2.fhir.fhirencountermapping import FHIREncounterMapping
from i2fhirb2.fhir.fhirnamespaces import fhir_namespace_for
from i2fhirb2.fhir.fhirpatientdimension import FHIRPatientDimension
from i2fhirb2.fhir.fhirpatientmapping import FHIRPatientMapping
from i2fhirb2.fhir.fhirprimitivetypes import i2b2_primitive, I2B2Value
from i2fhirb2.fhir.fhirproviderdimension import FHIRProviderDimension
from i2fhirb2.fhir.fhirspecific import composite_uri, instance_is_primitive
from i2fhirb2.fhir.fhirvisitdimension import FHIRVisitDimension
from i2fhirb2.i2b2model.data.i2b2observationfact import ObservationFact, ObservationFactKey
from i2fhirb2.sqlsupport.dynobject import DynElements


# TODO: Same patient_ide/patient_ide_src + visit_ide / visit_ide_src == duplicate


class FHIRObservationFact(ObservationFact):

    _unknown_namespaces: List[str] = []         # URI's that have been reported as being unknown

    def __init__(self, g: Graph, ofk: ObservationFactKey, concept: Union[URIRef, str],
                 modifier: Optional[Union[URIRef, str]], obj: Optional[Node],
                 instance_num: Optional[int] = None) -> None:
        """
        Construct an observation fact entry
        :param g: Graph containing information about the object
        :param ofk: Observation fact key - patient, encounter, etc
        :param concept: concept URI
        :param modifier: modifier URI (if any)
        :param obj: object of concept_code
        :param instance_num: instance identifier
        """
        super().__init__(ofk, self.ns_name_for(concept) if isinstance(concept, URIRef) else concept)
        if modifier is not None:
            self._modifier_cd = self.ns_name_for(modifier)
        self._instance_num = instance_num
        if obj is not None:
            v: I2B2Value = i2b2_primitive(g.value(obj, FHIR.value, any=False))
            self._valtype_cd = v.valtype
            self._tval_char = v.tval_char
            self._nval_num = v.nval_num
            self._observation_blob = v.observation_blob
        self.identifying_codes: List[str] = []

    @classmethod
    def ns_name_for(cls, concept_uri: Union[URIRef, str]) -> Optional[str]:
        """ Convert concept_uri into a NS:code construct if the namespace is known

        :param concept_uri: URI to convert
        :return: NS:code representation if NS is known, else None
        """
        concept_uri = str(concept_uri)
        if '#' in concept_uri:
            nsuri, cd = concept_uri.rsplit('#', 1)
            nsuri += '#'
        else:
            nsuri, cd = concept_uri.rsplit('/', 1)
            nsuri += '/'
        ns = fhir_namespace_for(nsuri)
        if ns is None:
            if nsuri not in cls._unknown_namespaces:
                cls._unknown_namespaces.append(nsuri)
                print(f"----> Unrecognized namespace: {nsuri}\n")
            return None

        ns = ns.upper()
        # Some FHIR resources seem to include the namespace as part of the code (?)
        return cd if cd.startswith(ns + ':') else ns.upper() + ':' + cd

    @classmethod
    def _clear(cls, complete=True):
        cls._unknown_namespaces = []
        ObservationFact._clear(complete)


class FHIRObservationFactFactory:
    """
    FHIRObservationFactFactory takes an RDF graph and generates a set of observation_fact, patient_dimension,
    visit_dimension, provider_dimension, patient_mapping and encounter_mapping entries
    """

    # List of predicates to be ignored in primary node generation
    # TODO: FHIR.Observation.referenceRange predicates are a temporary fix to issue 7.  Get a real fix in
    skip_predicates = [RDF.type, FHIR.nodeRole, FHIR.index, FHIR.link, FHIR.Observation.referenceRange]

    def __init__(self, g: Graph, ofk: ObservationFactKey, subject: Optional[URIRef]) -> None:
        self.g = g
        self.ofk = ofk
        self.observation_facts: List[FHIRObservationFact] = []
        self.patient_dimensions: List[FHIRPatientDimension] = []
        self.provider_dimensions: List[FHIRProviderDimension] = []
        self.visit_dimensions: List[FHIRVisitDimension] = []
        self.patient_mappings: List[FHIRPatientMapping] = []
        self.encounter_mappings: List[FHIREncounterMapping] = []
        self._inst_num = 0

        # TODO: look at DomainResource embedded entries (claim-example-oral-bridge).  Perhaps we should change the
        #       RDF generator to add type arcs to all resources?
        for s in {subject} if subject else self.g.subjects(FHIR.nodeRole, FHIR.treeRoot):
            self.observation_facts += self.generate_facts(s)

    def generate_facts(self, subject: URIRef) -> List[FHIRObservationFact]:
        """
        Generate an observation fact entry for each predicate object associated with subject.  If the predicate
        is not primitive (i.e. it has a non-value BNode as a target), generate a series of modifiers for each element
        in the BNode.  Note -- this generation sorts the inputs because the order is indirectly preserved in the output
        instance numbers.

        :param subject: Resource subject
        :return: List of observation facts for subject
        """
        rval: List[FHIRObservationFact] = []            # List of observation facts to add
        identifying_codes: List[str] = []               # Identifying code(s) for this type of entry
        for conc, obj in sorted(self.g.predicate_objects(subject)):
            if conc not in self.skip_predicates:
                coded_facts = process_concept_code(self.g, subject, subject, conc, self.ofk)
                if coded_facts:
                    identifying_codes = [e.concept_cd for e in coded_facts]
                    rval += coded_facts

                if instance_is_primitive(self.g, obj):
                    # The presence of a type arc on a primitive node indicates a code -- the value will be recorded
                    # elsewhere
                    if not self.g.value(obj, RDF.type):
                        rval.append(FHIRObservationFact(self.g, self.ofk, conc, None, obj))
                else:
                    rval += self.generate_modifiers(subject, conc, obj, identifying_codes=identifying_codes)
            elif conc == RDF.type:
                rval.append(FHIRObservationFact(self.g, self.ofk, obj, None, None))

        return self.removeduplicates(rval)

    @staticmethod
    def removeduplicates(facts: List[FHIRObservationFact]) -> List[FHIRObservationFact]:
        """ The process of adding modifiers can result in more than one subject concept_cd for the same value.  As an
        example, visionprescription-example-1.ttl.html has two products that are dispensed, both of which are contacts.

        :param facts: List of facts to be pruned
        :return: List with duplicates removed
        """
        fact_keys: List[Tuple[int, str, str]] = []
        rval: List[FHIRObservationFact] = []
        for fact in facts:
            k = (fact.instance_num, fact.concept_cd, fact.modifier_cd)
            if k not in fact_keys:
                fact_keys.append(k)
                rval.append(fact)
        return rval

    def generate_modifiers(self, subject: URIRef, concept: URIRef, obj: BNode,
                           parent_inst_num: int=0, same_subject: bool=False,
                           identifying_codes: Optional[List[str]]=None) -> List[FHIRObservationFact]:
        """
        Emit any modifiers for subject/obj.  If there is fhir:index field, this is an indication that this set
        of modifiers can occur multiple times and, as such, must be clustered.  Noting that, at the moment, we
        can only do one level of clustering, we do not generate unique instance numbers for singleton elements.

        :param subject: Resource subject
        :param concept: Concept identifier (first level entry)
        :param obj: inner cluster (usually a BNODE)
        :param parent_inst_num: instance number passed in recursive call
        :param same_subject: True means same parent -- don't revisit instance number
        :param identifying_codes: Code(s) assigned to the element
        :return:
        """
        rval = []               # type: List[FHIRObservationFact]
        fhir_index = self.g.value(obj, FHIR.index, any=False) if not same_subject else None
        pred_obj_list = \
            sorted([(p, o) for p, o in self.g.predicate_objects(obj) if p not in self.skip_predicates])

        if fhir_index is not None and len(pred_obj_list):
            self._inst_num += 1
            inst_num = self._inst_num
        else:
            inst_num = parent_inst_num

        for modifier, modifier_object in pred_obj_list:
            if instance_is_primitive(self.g, modifier_object):
                mod_inst_idx = self.g.value(modifier_object, FHIR.index, any=False)
                if mod_inst_idx is not None and len(pred_obj_list) > 1:
                    self._inst_num += 1
                    inst_num = self._inst_num
                obsfact = FHIRObservationFact(self.g, self.ofk, concept, modifier, modifier_object, inst_num)
                rval.append(obsfact)
            else:
                rval += self.generate_modifiers(subject,
                                                concept if inst_num else composite_uri(concept, modifier),
                                                modifier_object,
                                                inst_num,
                                                same_subject=True,
                                                identifying_codes=identifying_codes)
            rval += process_concept_code(self.g, concept, obj, modifier, self.ofk, inst_num, identifying_codes)
        return rval



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

# create table i2b2demodata.patient_dimension
# (
# 	patient_num integer not null
# 		constraint patient_dimension_pk
# 			primary key,
# 	vital_status_cd varchar(50),
# 	birth_date timestamp,
# 	death_date timestamp,
# 	sex_cd varchar(50),
# 	age_in_years_num integer,
# 	language_cd varchar(50),
# 	race_cd varchar(50),
# 	marital_status_cd varchar(50),
# 	religion_cd varchar(50),
# 	zip_cd varchar(10),
# 	statecityzip_path varchar(700),
# 	income_cd varchar(50),
# 	patient_blob text,
# 	update_date timestamp,
# 	download_date timestamp,
# 	import_date timestamp,
# 	sourcesystem_cd varchar(50),
# 	upload_id integer
# )
from datetime import datetime
from typing import Optional, Tuple, List

from i2fhirb2.i2b2model.shared.i2b2core import I2B2CoreWithUploadId
from i2fhirb2.sqlsupport.dynobject import DynElements, DynObject
from i2fhirb2.sqlsupport.dbconnection import I2B2Tables


class VitalStatusCd:
    def __setattr__(self, key, value):
        if key not in ["birthcode", "deathcode"]:
            raise ValueError("New elements not allowed")
        super().__setattr__(key, value)

    class BirthDateCode:
        def __init__(self, code: str) -> None:
            self.code = code

    bd_unknown = BirthDateCode('L')
    bd_day = BirthDateCode('D')
    bd_month = BirthDateCode('B')
    bd_year = BirthDateCode('F')
    bd_hour = BirthDateCode('H')
    bd_minute = BirthDateCode('M')
    bd_second = BirthDateCode('C')

    class DeathDateCode:
        def __init__(self, code: str) -> None:
            self.code = code
    dd_living = DeathDateCode('N')
    dd_unknown = DeathDateCode('U')
    dd_deceased = DeathDateCode('Z')
    dd_day = DeathDateCode('Y')
    dd_month = DeathDateCode('M')
    dd_year = DeathDateCode('X')
    dd_hour = DeathDateCode('R')
    dd_minute = DeathDateCode('T')
    dd_second = DeathDateCode('S')

    def __init__(self, birth: BirthDateCode, death: DeathDateCode) -> None:
        self.birthcode = birth
        self.deathcode = death

    @property
    def code(self):
        return self.deathcode.code + self.birthcode.code


# TODO: should age be computed from birthdate / deathdate
# TODO: language code -- what do we do with this?

class PatientDimension(I2B2CoreWithUploadId):
    _t = DynElements(I2B2CoreWithUploadId)

    key_fields = ["patient_num"]

    def __init__(self, patient_num, vital_status_cd: VitalStatusCd, **kwargs) -> None:
        self._patient_num = patient_num
        self._vital_status_code = vital_status_cd
        self._birth_date = None
        self._death_date = None
        self._sex_cd = None
        self._age_in_years_num = None
        self._language_cd = None
        self._race_cd = None
        self._marital_status_cd = None
        self._religion_cd = None
        self._zip_cd = None
        self._statecityzip_path = None
        self._income_cd = None
        self._patient_blob = None
        super().__init__(**kwargs)

    @DynObject.entry(_t)
    def patient_num(self) -> int:
        """
        Reference number for the patient
        """
        return self._patient_num

    @DynObject.entry(_t)
    def vital_status_cd(self) -> Optional[str]:
        """
        Encoded i2b2 patient visit number
        """
        return self._vital_status_code.code

    @DynObject.entry(_t)
    def birth_date(self) -> Optional[datetime]:
        """
        Encoded i2b2 patient visit number
        """
        return self._birth_date

    @DynObject.entry(_t)
    def death_date(self) -> Optional[datetime]:
        """
        Encoded i2b2 patient visit number
        """
        return self._death_date

    @DynObject.entry(_t)
    def sex_cd(self) -> Optional[str]:
        """
        Encoded i2b2 patient visit number
        """
        return self._sex_cd

    @DynObject.entry(_t)
    def age_in_years_num(self) -> Optional[int]:
        """
        Encoded i2b2 patient visit number
        """
        return self._age_in_years_num
    
    @DynObject.entry(_t)
    def language_cd(self) -> Optional[str]:
        return self._language_cd

    @DynObject.entry(_t)
    def race_cd(self) -> Optional[str]:
        """
        Encoded i2b2 patient visit number
        """
        return self._race_cd

    @DynObject.entry(_t)
    def marital_status_cd(self) -> Optional[str]:
        return self._marital_status_cd
    
    @DynObject.entry(_t)
    def religion_cd(self) -> Optional[str]:
        return self._religion_cd
    
    @DynObject.entry(_t)
    def zip_cd(self) -> Optional[str]:
        return self._zip_cd
    
    @DynObject.entry(_t)
    def statecityzip_path(self) -> Optional[str]:
        return self._statecityzip_path
    
    @DynObject.entry(_t)
    def income_cd(self) -> Optional[str]:
        return self._income_cd
    
    @DynObject.entry(_t)
    def patient_blob(self) -> Optional[str]:
        return self._patient_blob

    @classmethod
    def delete_upload_id(cls, tables: I2B2Tables, upload_id: int) -> int:
        """
        Delete all patient_dimension records with the supplied upload_id
        :param tables: i2b2 sql connection
        :param upload_id: upload identifier to remove
        :return: number or records that were deleted
        """
        return cls._delete_upload_id(tables.crc_connection, tables.patient_dimension, upload_id)

    @classmethod
    def delete_sourcesystem_cd(cls, tables: I2B2Tables, sourcesystem_cd: str) -> int:
        """
        Delete all records with the supplied sourcesystem_cd
        :param tables: i2b2 sql connection
        :param sourcesystem_cd: sourcesystem_cd to remove
        :return: number or records that were deleted
        """
        return cls._delete_sourcesystem_cd(tables.crc_connection, tables.patient_dimension, sourcesystem_cd)

    @classmethod
    def add_or_update_records(cls, tables: I2B2Tables, records: List["PatientDimension"]) -> Tuple[int, int]:
        """
        Add or update the patient_dimension table as needed to reflect the contents of records
        :param tables: i2b2 sql connection
        :param records: records to apply
        :return: number of records added / modified
        """
        return cls._add_or_update_records(tables.crc_connection, tables.patient_dimension, records)



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
import os
from typing import List


def esc_output(txt: str) -> str:
    """
    Escape carriage returns for tsv output
    :param txt:
    :return:
    """
    return txt.replace('\r\n', '').replace('\r', '').replace('\n', '')


def write_tsv(filedir: str, file: str, hdr: str, values: List[object]) -> bool:
    ofn = filedir + file + ('.tsv' if '.' not in file else '')
    os.makedirs(os.path.dirname(ofn), exist_ok=True)
    print("writing {}".format(ofn), end="")
    with open(ofn, 'w') as outf:
        outf.write(hdr + '\n')
        for e in sorted(values):
            outf.write(esc_output(repr(e)) + '\n')
    print(" ({}) records written".format(len(values)))
    return True



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
import argparse
from argparse import Namespace
from typing import Callable, Optional

from i2fhirb2.file_aware_parser import FileAwareParser
from i2fhirb2.i2b2model.shared.tablenames import i2b2tablenames, DEFAULT_ONTOLOGY_TABLE
from i2fhirb2.sqlsupport.i2b2tables import I2B2Tables


class ConfigFile(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs) -> None:
        super().__init__(option_strings, dest, nargs, **kwargs)

    def __call__(self, parser: FileAwareParser, namespace: Namespace, values, option_string=None):
        raise AttributeError("Must preprocess input arguments with decode_file_args function")


def add_connection_args(parser: FileAwareParser, strong_config_file: bool=True) -> FileAwareParser:
    """
    Add the database connection arguments to the supplied parser

    :param parser: parser to add arguments to
    :param strong_config_file: If True, force --conf to be processed.  This is strictly a test for programming errors,
      and has to be skipped due to removefacts function.
    :return: parser
    """
    # TODO: Decide what to do with this
    parser.add_file_argument("--conf", metavar="CONFIG FILE", help="Configuration file",
                             action=ConfigFile if strong_config_file else None)

    parser.add_argument("-db", "--dburl", help="Default database URL",
                        default=Default_DB_Connection)
    parser.add_argument("--user", help="Default user name",
                        default=Default_User)
    parser.add_argument("--password", help="Default password",
                        default=Default_Password)
    parser.add_argument("--crcdb", help="CRC database URL. (default: dburl)")
    parser.add_argument("--crcuser", help="User name for CRC database. (default: user)")
    parser.add_argument("--crcpassword", help="Password for CRC database. (default: password)")
    parser.add_argument("--ontodb", help="Ontology database URL.  (default: dburl)")
    parser.add_argument("--ontouser", help="User name for ontology database. (default: user)")
    parser.add_argument("--ontopassword", help="Password for ontology database. (default: password)")
    parser.add_argument("--onttable", metavar="ONTOLOGY TABLE NAME",
                        help="Ontology table name (default: {})".format(DEFAULT_ONTOLOGY_TABLE),
                        default=DEFAULT_ONTOLOGY_TABLE)
    return parser


def process_parsed_args(opts: Namespace, error_fun: Optional[Callable], connect: bool=True) -> Namespace:
    """
    Set the defaults for the crc and ontology schemas
    :param opts: parsed arguments
    :param error_fun: Function to call if error
    :param connect: actually connect. (For debugging)
    :return: namespace with additional elements added
    """
    def setdefault(vn: str, default: object) -> None:
        assert vn in opts, "Unknown option"
        if not getattr(opts, vn):
            setattr(opts, vn, default)

    if error_fun and \
            (getattr(opts, 'dburl') is None or getattr(opts, 'user') is None or getattr(opts, 'password') is None):
        error_fun("db url, user id and password must be supplied")
    setdefault('crcdb', opts.dburl)
    setdefault('crcuser', opts.user)
    setdefault('crcpassword', opts.password)
    setdefault('ontodb', opts.dburl)
    setdefault('ontouser', opts.user)
    setdefault('ontopassword', opts.password)
    if connect:
        opts.tables = I2B2Tables(opts)

    # TODO: This approach needs to be re-thought.  As i2b2tablenames is a singleton, any changes here
    # impact the entire testing harness
    if opts.onttable:
        i2b2tablenames.ontology_table = opts.onttable
    return opts


Default_DB_Connection = "postgresql+psycopg2://localhost:5432/i2b2"
Default_User = "postgres"
Default_Password = "postgres"


# Copyright (c) 2018, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from datetime import datetime
from typing import Optional, Dict, Callable

from rdflib import Literal, URIRef, XSD

from i2fhirb2.i2b2model.data.i2b2observationfact import ValueTypeCd, valuetype_text, valuetype_blob, valuetype_date, \
    valuetype_number


class I2B2Value:
    """ Representation of th Value portion of an I2B2 Observation Fact"""
    def __init__(self, tval_char: str, valtype: ValueTypeCd= valuetype_text, nval_num: Optional[float]=None,
                 observation_blob: Optional[str]=None) -> None:
        self.tval_char = tval_char
        self.valtype = valtype
        self.nval_num = nval_num
        self.observation_blob = observation_blob

    def __repr__(self) -> str:
        bval = f"{self.observation_blob[:10]}...{self.observation_blob[-10:]}" \
            if self.observation_blob and len(self.observation_blob) > 20 \
            else self.observation_blob
        return f'I2B2Value("{self.tval_char}", ValueTypeCd("{self.valtype.code}"), {self.nval_num}, {bval})'

    def __eq__(self, other: "I2B2Value"):
        return repr(self) == repr(other)


def filter_literal(val: Literal) -> str:
    return format(str(val).replace('\n', '\\n').replace(r'\t', '\\t'))


def blob_val(val: Literal) -> I2B2Value:
    return I2B2Value("", valuetype_blob, observation_blob='"' + filter_literal(val) + '"')


def text_val(val: Literal) -> I2B2Value:
    return I2B2Value(filter_literal(val))


def date_val(val: Literal) -> I2B2Value:
    dt = val.value
    nval_num = (dt.year * 10000) + (dt.month * 100) + dt.day + \
               (((dt.hour / 100.0) + (dt.minute / 10000.0)) if isinstance(dt, datetime) else 0)
    return I2B2Value(dt.strftime('%Y-%m-%d %H:%M'), valuetype_date, nval_num)


def decimal_val(val: Literal) -> I2B2Value:
    return I2B2Value("E", valuetype_number, float(val.value))


def time_val(val: Literal) -> I2B2Value:
    # TODO: Proposed format.  Verify w/ i2b2 community
    tv = val.value
    nval_num = (tv.hour / 100.0) + (tv.minute / 10000.0) + (tv.second / 1000000.0)
    return I2B2Value(str(val.value), valuetype_date, nval_num)


# Conversion table from XSD data type to corresponding i2b2 field
# TODO: time has no analog in i2b2 - should it?
literal_conversions: Dict[URIRef, Callable[[Literal], I2B2Value]] = {
    XSD.base64Binary: blob_val,
    XSD.boolean: text_val,
    XSD.date: date_val,
    XSD.dateTime: date_val,
    XSD.decimal: decimal_val,
    XSD.gYear: date_val,
    XSD.gYearMonth: date_val,
    XSD.integer: decimal_val,
    XSD.nonNegativeInteger: decimal_val,
    XSD.positiveInteger: decimal_val,
    XSD.time: time_val
}


def i2b2_primitive(val: Literal) -> I2B2Value:
    return literal_conversions[val.datatype](val) if val.datatype and val.datatype in literal_conversions \
        else text_val(val)



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Set, List, Optional

from fhirtordf.rdfsupport.namespaces import W5
from rdflib import Graph, URIRef, RDF, RDFS, OWL

from i2fhirb2.common_cli_parameters import DEFAULT_NAME_BASE
from i2fhirb2.fhir.fhirspecific import w5_infrastructure_categories, concept_path

INDENT = '  '


def is_w5_uri(uri: URIRef) -> bool:
    """ Determine whether uri is in the w5 namespace

    :param uri: URI to test
    :return: True if in w5 namespace
    """
    return str(uri).startswith(str(W5))


class W5GraphNode:
    """
    A node and all of the descendants down to and including the first non-w5 entry
    """
    def __init__(self, g: Graph, node: URIRef) -> None:
        self.node = node
        self.children = [W5GraphNode(g, subj) for subj in g.subjects(RDFS.subClassOf, node)
                         if isinstance(subj, URIRef)] if is_w5_uri(node) else []

    def indented_str(self, indent: int) -> str:
        rval = INDENT * indent + ('Node: {}' if is_w5_uri(self.node) else 'Resource: {}').format(self.node)
        if self.children:
            rval += '\n' + '\n'.join([c.indented_str(indent + 1) for c in sorted(self.children)])
        return rval

    def __str__(self) -> str:
        return self.indented_str(0)

    def __lt__(self, other: "W5GraphNode") -> bool:
        return self.node < other.node

    def __eq__(self, other: "W5GraphNode") -> bool:
        return self.node == other.node


class W5PathEntry:
    def __init__(self, path: str, w5_ent: W5GraphNode) -> None:
        """ A w5 path and accompanying metadata

        :param path: Full path to referenced node
        :param w5_ent: Corresponding W5GraphNode
        """
        self.node = w5_ent.node
        self.path = path
        self.fhir_resource_uri = self.node if not is_w5_uri(w5_ent.node) else None

    def __lt__(self, other) -> bool:
        return self.path < other.path

    def __eq__(self, other):
        return self.path == other.path

    def __str__(self) -> str:
        return self.path + (":({})".format(self.fhir_resource_uri) if self.fhir_resource_uri else "")


class FHIRW5Ontology:
    def __init__(self, g: Graph) -> None:
        """ A representation of the FHIR W5 GRAPH

        :param g: Graph containing W5 concepts
        """
        self.g = g
        self.w5_graph = [W5GraphNode(g, subj) for subj in self._w5_concepts() if g.value(subj, RDFS.subClassOf) is None]

    def _w5_concepts(self) -> Set[URIRef]:
        """
        Return a list of all W5 classes that aren't infrastructure classes
        """
        return {subj for subj in self.g.subjects(RDF.type, OWL.Class)
                if isinstance(subj, URIRef) and is_w5_uri(subj) and not self.is_w5_infrastructure(subj)}

    def is_w5_infrastructure(self, subj: URIRef) -> bool:
        """
        Determine whether subj belongs to a w5 infrastructure category
        :param subj: FHIR Element
        :return: True if this is an infrastructure category, false otherwise
        """
        return subj in w5_infrastructure_categories or \
            bool(set(self.g.transitive_objects(subj, RDFS.subClassOf)).intersection(w5_infrastructure_categories))

    def _nested_path(self, base: str, w5_ent: W5GraphNode) -> List[W5PathEntry]:
        node_path = base + concept_path(w5_ent.node)
        rval = [W5PathEntry(node_path, w5_ent)]
        for child in w5_ent.children:
            rval += self._nested_path(node_path, child)
        return rval

    def w5_paths(self, name_base: Optional[str] = DEFAULT_NAME_BASE) -> List[W5PathEntry]:
        rval = []
        for w5_ent in self.w5_graph:
            rval += self._nested_path(name_base, w5_ent)
        return rval

    def __str__(self):
        return "W5 Ontology\n" + '\n'.join([c.indented_str(1) for c in sorted(self.w5_graph)])



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Optional

from i2fhirb2.i2b2model.metadata.i2b2ontologyquery import ConceptQuery
from i2fhirb2.i2b2model.metadata.i2b2ontologyvisualattributes import VisualAttributes
from i2fhirb2.i2b2model.shared.tablenames import i2b2tablenames
from i2fhirb2.sqlsupport.dynobject import DynObject, DynElements


class TableAccess(DynObject):
    _t = DynElements(DynObject)
    _visualattributes = VisualAttributes("CA")
    _query = ConceptQuery("\\FHIR\\")

    @DynObject.entry(_t)
    def c_table_cd(self) -> str:
        return "FHIR"

    @DynObject.entry(_t)
    def c_table_name(self) -> str:
        return i2b2tablenames.phys_name(i2b2tablenames.ontology_table)

    @DynObject.entry(_t)
    def c_protected_access(self) -> str:
        return "N"

    @DynObject.entry(_t)
    def c_hlevel(self) -> int:
        return 1

    @DynObject.entry(_t)
    def c_fullname(self) -> str:
        return '\\FHIR\\'

    @DynObject.entry(_t)
    def c_name(self) -> str:
        return "FHIR Resources"

    @DynObject.entry(_t)
    def c_synonym_cd(self) -> str:
        return "N"

    @DynObject.entry(_t)
    def c_visualattributes(self) -> str:
        return str(self._visualattributes)

    @DynObject.entry(_t)
    def c_totalnum(self) -> Optional[int]:
        return None

    @DynObject.entry(_t)
    def c_basecode(self) -> Optional[str]:
        return None

    @DynObject.entry(_t)
    def c_metadataxml(self) -> Optional[str]:
        return None

    @DynObject.entry(_t)
    def c_facttablecolumn(self) -> str:
        return self._query.key

    @DynObject.entry(_t)
    def c_dimtablename(self) -> str:
        return self._query.table

    @DynObject.entry(_t)
    def c_columnname(self) -> str:
        return self._query.where_subj

    @DynObject.entry(_t)
    def c_columndatatype(self) -> str:
        return 'N' if self._query.numeric_key else 'T'

    @DynObject.entry(_t)
    def c_operator(self) -> str:
        return self._query.where_pred

    @DynObject.entry(_t)
    def c_dimcode(self) -> str:
        return self._query.where_obj

    @DynObject.entry(_t)
    def c_comment(self) -> Optional[str]:
        return None

    @DynObject.entry(_t)
    def c_tooltip(self) -> Optional[str]:
        return "FHIR Resource"

    @DynObject.entry(_t)
    def c_entry_date(self) -> Optional[str]:
        return None

    @DynObject.entry(_t)
    def c_change_date(self) -> Optional[str]:
        return None

    @DynObject.entry(_t)
    def c_status_cd(self) -> Optional[str]:
        return None

    @DynObject.entry(_t)
    def valuetype_cd(self) -> Optional[str]:
        return None



# Copyright (c) 2018, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
import copy
from typing import List, Dict, Callable, Optional

from fhirtordf.rdfsupport import fhirgraphutils
from fhirtordf.rdfsupport.namespaces import FHIR
from rdflib import URIRef, Graph, RDF
from rdflib.term import Node, Literal

from i2fhirb2.fhir.fhirnamespaces import fhir_namespace_for
from i2fhirb2.i2b2model.data.i2b2observationfact import ObservationFactKey, ObservationFact, valuetype_text, \
    valuetype_number, valuetype_novalue


def value_string(g: Graph, node: Node, obs_fact: ObservationFact) -> List[ObservationFact]:
    """ Convert a FHIR.string type value into the i2b2 equivalent

    :param g: Graph containing the data
    :param node: Node containing the string value
    :param obs_fact: target fact to have string added to it
    :return: Empty list - no additional facts are generated
    """
    obs_fact._valtype_cd = valuetype_text
    obs_fact._tval_char = g.value(node, FHIR.value, default=Literal(""), any=False).value
    return []


comparator_map = {'<': 'L',
                  '<=': 'LE',
                  '>=': 'GE',
                  '>': 'G'}


def value_quantity(g: Graph, subject: Node, obs_fact: ObservationFact) -> List[ObservationFact]:
    """ Process a FHIR:valueQuantity, recording salient information in obs_fact

    :param g: Graph containing the quantity
    :param subject: Subject of the fhir quantity node
    :param obs_fact: Fact to add the information to
    :return: Empty list - no additional facts are generated
    """
    units = fhirgraphutils.value(g, subject, FHIR.Quantity.unit)
    if not units:
        system = fhirgraphutils.value(g, subject, FHIR.Quantity.system)
        code = fhirgraphutils.value(g, subject, FHIR.Quantity.code)
        if code:
            ns = fhir_namespace_for(system)
            units = ((ns.upper() + ':') if ns else '') + code

    comparator = fhirgraphutils.value(g, subject, FHIR.Quantity.comparator)
    value_ = fhirgraphutils.value(g, subject, FHIR.Quantity.value)
    if value_ is not None:
        obs_fact._valtype_cd = valuetype_number
        obs_fact._nval_num = value_
        obs_fact._tval_char = 'E' if comparator is None else comparator_map.get(str(comparator), '?')
        obs_fact._units_cd = str(units) if units else None
    else:
        obs_fact._valtype_cd = valuetype_novalue
    return []


def value_integer(g: Graph, subject: Node, obs_fact: ObservationFact) -> List[ObservationFact]:
    """ Process a FHIR:valueInteger, recording salient information in obs_fact

    :param g: Graph containing the quantity
    :param subject: Subject of the fhir quantity node
    :param obs_fact: Fact to add the information to
    :return: Empty list - no additional facts are generated
    """
    obs_fact._valtype_cd = valuetype_number
    obs_fact._nval_num = g.value(subject, FHIR.value, any=False).value
    obs_fact._tval_char = 'E'
    return []


def value_codeable_concept(g: Graph, subject: Node, obs_fact: ObservationFact) -> List[ObservationFact]:
    """ Process a FHIR:valueCodeableConcept node.  The first FHIR:CodeableConcept.coding entry is recorded as
    the value for the passed `obs_fact`.  Additional obs_fact entries are generated to carry the text, if it exists,
    and any additional codings.

    :param g: Graph containing the data
    :param subject: CodeableConcept node
    :param obs_fact: target fact(s)
    :return: Additional obs_facts beyond obs_fact itself
    """
    from i2fhirb2.fhir.fhirobservationfact import FHIRObservationFact

    rval = []
    if obs_fact.modifier_cd != '@':
        print(f"{obs_fact.concept_cd}, {obs_fact.modifier_cd}: Modifier on modifier!")
    codings = sorted(list(g.objects(subject, FHIR.CodeableConcept.coding)))
    additional_entries = False
    if codings:
        for coding in codings:
            obs_fact_copy = copy.copy(obs_fact)
            if obs_fact.modifier_cd == '@':
                concept_uri = concept_uri_for(g, coding)
                concept_ns_name = FHIRObservationFact.ns_name_for(concept_uri) if concept_uri else None
                if concept_ns_name:
                    obs_fact._modifier_cd = concept_ns_name
            display = fhirgraphutils.value(g, coding, FHIR.Coding.display)
            if not display:
                display = fhirgraphutils.value(g, subject, FHIR.CodeableConcept.text)
            if display:
                obs_fact._valtype_cd = valuetype_text
                obs_fact._tval_char = display
                if additional_entries:
                    rval.append(obs_fact)
                else:
                    additional_entries = True
                obs_fact = obs_fact_copy
    else:
        text = fhirgraphutils.value(g, subject, FHIR.CodeableConcept.text)
        if text:
            obs_fact._valtype_cd = valuetype_text
            obs_fact._tval_char = text

    return rval


# Table of value types to i2b2 converters
value_processors: Dict[str, Callable[[Graph, Node, ObservationFact], List[ObservationFact]]] = {
    "valueQuantity": value_quantity,
    "valueCodeableConcept": value_codeable_concept,
    "valueString": value_string,
    "valueBoolean": None,
    "valueInteger": value_integer,
    "valueRange": None,
    "valueRatio": None,
    "valueSampledData": None,
    "valueTime": None,
    "valueDateTime": None,
    "valuePeriod": None
}


def proc_value_node(g: Graph, obs_fact: ObservationFact, predicate: URIRef, object: Node) -> List[ObservationFact]:
    """ Process a value node, adding the value elements to obs_fact and returning a list of any additional facts
    that get generated in the process

    :param g:
    :param obs_fact: The primary fact to have values added to it
    :param predicate: The predicate (type) of value
    :param object: The target resource
    :return: Any additional observation facts *beyond* obs_fact
    """

    # TODO: Interpretation
    value_type = str(predicate).split('.')[-1]
    if value_type not in value_processors:
        raise NotImplementedError(f"Unrecognized value type: {value_type}")
    if value_processors[value_type] is not None:
        return value_processors[value_type](g, object, obs_fact)
    else:
        print(f"***> Unimplemented FHIR value: {value_type}")

    return []


def is_fhir_value(p: URIRef) -> bool:
    """ Determine whether `p` is a fhir:value[x] node

    :param p: predicate to test
    :return:
    """
    return str(p).startswith(str(FHIR)) and str(p).split('.')[-1].startswith('value')


def concept_uri_for(g: Graph, coding: Node) -> Optional[URIRef]:
    """ Return the URI for the concept in coding, if any.  If a type arc is present, it will be used.  Otherwise
    we will attempt to synthesize one from system and code.

    :param g: Graph containing the coding node
    :param coding: node that contains the FHIR.Coding entry
    :return: URI or None if nothing is available
    """
    uri = g.value(coding, RDF.type)
    if uri is None:
        # TODO: Combine this with the rules in fhirtordf - they should do exactly the same thing
        system = fhirgraphutils.value(g, coding, FHIR.Coding.system)
        code = fhirgraphutils.value(g, coding, FHIR.Coding.code)
        if system and code:
            uri = URIRef(system + ('' if system.endswith(('#', '/')) else '/') + code)
    return uri


def add_concept_values(g: Graph, obs_fact: ObservationFact, subject: URIRef, graph_node: Node) -> List[ObservationFact]:
    """  Add any value[x] entries to obs_fact

    :param g:
    :param obs_fact: Base fact
    :param subject: Subject of values
    :param graph_node: Node that may have value node predicates
    :return: Fact(s) with values added
    """
    rval = [obs_fact]
    if subject in code_value_tuples:
        # Add any associated value(s)
        for p, o in g.predicate_objects(graph_node):
            if is_fhir_value(p):
                rval += proc_value_node(g, obs_fact, p, o)
    elif str(subject).split('.')[-1] == "valueCodeableConcept":
        for o in g.objects(graph_node, subject):
            rval += proc_value_node(g, obs_fact, subject, o)
    return rval


def process_concept_code(g: Graph, root_concept: URIRef, subject: Node, predicate: URIRef, ofk: ObservationFactKey,
                         instance_num: Optional[int]=None, identifying_codes: Optional[List[str]]=None) \
        -> List[ObservationFact]:
    """ If predicate meets the requirements, emit additional facts that incorporate the referenced coded concept, using
    the following rules:
    1) If predicate occurs at the root level of a FHIR Resource:
        If predicate is marked as identifying, emit the referenced concept as a concept_cd
        Otherwise, emite predicate as the concept_cd and the referenced concept as a modifier_cd
    2) If the predicate occurs as a non-root (instance_num is not None), add the following entries:
            1) concept_cd = concept_code_coding, modifier_cd = '@' inst_num = 0
            2) concept_cd = identifying_code, modifier_cd=concept_code_coding, inst_num = 0
            3) concept_cd = predicate, modifier_cd=concept_code_coding, inst_num = inst_num

    :param g: Graph containing facts
    :param root_concept: level 0 concept associated with this entry
    :param subject: subject
    :param predicate: predicate
    :param ofk: Observation fact key
    :param instance_num: Observation fact instance number if non-zero
    :param identifying_codes: Observation fact root codes if inside an instance
    :return: Observation fact(s) if any that augment the basic subject/predicate
    """
    from i2fhirb2.fhir.fhirobservationfact import FHIRObservationFact   # Imported here to prevent recursive imports

    rval: List[ObservationFact] = []

    # Find any codeable concepts in the range of predicate
    for target in g.objects(subject, predicate):                        # Indirectly looking for FHIR:CodedEntry
        for target_p, target_o in g.predicate_objects(target):
            if target_p == FHIR.CodeableConcept.coding:                 # And interior FHIR:Coding
                target_concept_code = concept_uri_for(g, target_o)      # Look for type arc or synthesize it
                # Code must exist and we have to have a known namespace to proceed
                if target_concept_code and FHIRObservationFact.ns_name_for(target_concept_code):
                    if predicate in bare_codes:
                        # Predicate can serve as a concept_cd for the resource itself
                        # Add the root element and its values
                        rval += add_concept_values(g,
                                                   FHIRObservationFact(g, ofk, target_concept_code, None, None),
                                                   predicate,
                                                   subject)
                    if instance_num is None:
                        # Add an entry as a concept/modifier entry
                        rval += add_concept_values(g,
                                                   FHIRObservationFact(g, ofk, predicate, target_concept_code, None),
                                                   predicate,
                                                   subject)

                    else:
                        # Inner coded element -- create a set of base entries
                        base_entries = add_concept_values(
                            g,
                            FHIRObservationFact(g, ofk, root_concept, target_concept_code, None, instance_num),
                            predicate,
                            subject)
                        rval += base_entries
                        for id_code in identifying_codes:
                            # TODO: Kludge - id_code is in ns:name form and identifying codes are in URI form.  For the
                            # time being we are going to assume that the FHIR coding system will not be used for
                            # secondary identity.  Fix this
                            # if id_code not in bare_codes:
                            if not id_code.startswith("FHIR:"):
                                for entry in base_entries:
                                    fact_entry = copy.copy(entry)
                                    fact_entry._concept_cd = id_code
                                    fact_entry._modifier_cd = FHIRObservationFact.ns_name_for(target_concept_code)
                                    fact_entry._instance_num = 0
                                    rval.append(fact_entry)
                                if not base_entries:
                                    rval.append(FHIRObservationFact(g, ofk, id_code, target_concept_code, None))
            elif target_p == RDF.type and predicate not in (FHIR.CodeableConcept.coding, ):
                # For the moment, we assume that the only other type nodes are for ``code`` nodes
                # Code must exist and we have to have a known namespace to proceed
                if FHIRObservationFact.ns_name_for(target_o):
                    cvals = add_concept_values(g,
                                               FHIRObservationFact(g, ofk, predicate, target_o, None, None),
                                               predicate,
                                               subject)
                    cval = cvals[0]
                    cval._valtype_cd = valuetype_text
                    cval._tval_char = str(g.value(target, FHIR.value))
                    rval += cvals
    return rval


# List of concept codes that serve as identifier codes, meaning that they become standalone
# codes for the root node.
bare_codes: List[URIRef] = {
    FHIR.AdverseEvent.event,
    FHIR.AdverseEvent.seriousness,
    FHIR.AdverseEvent.severity,
    FHIR.AdverseEvent.outcome,

    # TODO: What to do about reaction[0..*] manifestation[1..*] entries? + exposure route
    FHIR.AllergyIntolerance.code,

    # CarePlan doesn't appear to have any primary elements.

    FHIR.Condition.code,

    FHIR.DiagnosticReport.code,

    FHIR.MedicationAdministration.medicationCodeableConcept,

    FHIR.MedicationDispense.medicationCodeableConcept,

    FHIR.MedicationRequest.medicationCodeableConcept,

    FHIR.MedicationStatement.medicationCodeableConcept,

    FHIR.NutritionOrder.oralDiet.type,
    FHIR.NutritionOrder.supplement.type,
    FHIR.NutritionOrder.enteralFormula.baseFormulaType,

    FHIR.Observation.code,
    FHIR.Observation.component.code,

    # TODO: We could consider folding eye into product where appropriate
    FHIR.VisionPrescription.dispense.product,           # Product that was dispensed
    FHIR.VisionPrescription.reasonCodeableConcept       # Indication for prescription
}

code_value_tuples: List[URIRef] = [
    FHIR.Observation.code,
    FHIR.Observation.component.code,
]



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Optional, List, Tuple

from i2fhirb2.i2b2model.shared.i2b2core import I2B2CoreWithUploadId
from i2fhirb2.sqlsupport.dynobject import DynElements, DynObject
from i2fhirb2.sqlsupport.dbconnection import I2B2Tables


class EncounterIDEStatus:
    class EncounterIDEStatusCode:
        def __init__(self, code: str) -> None:
            self.code = code
    active = EncounterIDEStatusCode("A")
    inactive = EncounterIDEStatusCode("I")
    deleted = EncounterIDEStatusCode("D")
    merged = EncounterIDEStatusCode("M")


class EncounterMapping(I2B2CoreWithUploadId):
    _t = DynElements(I2B2CoreWithUploadId)

    key_fields = ["encounter_ide", "encounter_ide_source", "project_id", "patient_ide", "patient_ide_source"]

    def __init__(self,
                 encounter_ide: str,
                 encounter_ide_source: str,
                 project_id: str,
                 encounter_num: int,
                 patient_ide: str,
                 patient_ide_source: str,
                 encounter_ide_status: Optional[EncounterIDEStatus.EncounterIDEStatusCode],
                 **kwargs):
        self._encounter_ide = encounter_ide
        self._encounter_ide_source = encounter_ide_source
        self._project_id = project_id
        self._encounter_num = encounter_num
        self._patient_ide = patient_ide
        self._patient_ide_source = patient_ide_source
        self._encounter_ide_status = encounter_ide_status

        super().__init__(**kwargs)

    @DynObject.entry(_t)
    def encounter_ide(self) -> str:
        return self._encounter_ide

    @DynObject.entry(_t)
    def encounter_ide_source(self) -> str:
        return self._encounter_ide_source

    @DynObject.entry(_t)
    def project_id(self) -> Optional[str]:
        return self._project_id

    @DynObject.entry(_t)
    def encounter_num(self) -> int:
        return self._encounter_num

    @DynObject.entry(_t)
    def patient_ide(self) -> str:
        return self._patient_ide

    @DynObject.entry(_t)
    def patient_ide_source(self) -> str:
        return self._patient_ide_source

    @DynObject.entry(_t)
    def encounter_ide_status(self) -> Optional[str]:
        return self._encounter_ide_status.code if self._encounter_ide_status else None

    @classmethod
    def delete_upload_id(cls, tables: I2B2Tables, upload_id: int) -> int:
        """
        Delete all patient_dimension records with the supplied upload_id
        :param tables: i2b2 sql connection
        :param upload_id: upload identifier to remove
        :return: number or records that were deleted
        """
        return cls._delete_upload_id(tables.crc_connection, tables.encounter_mapping, upload_id)

    @classmethod
    def delete_sourcesystem_cd(cls, tables: I2B2Tables, sourcesystem_cd: str) -> int:
        """
        Delete all records with the supplied sourcesystem_cd
        :param tables: i2b2 sql connection
        :param sourcesystem_cd: sourcesystem_cd to remove
        :return: number or records that were deleted
        """
        return cls._delete_sourcesystem_cd(tables.crc_connection, tables.encounter_mapping, sourcesystem_cd)

    @classmethod
    def add_or_update_records(cls, tables: I2B2Tables, records: List["EncounterMapping"]) -> Tuple[int, int]:
        """
        Add or update the patient_dimension table as needed to reflect the contents of records
        :param tables: i2b2 sql connection
        :param records: records to apply
        :return: number of records added / modified
        """
        return cls._add_or_update_records(tables.crc_connection, tables.encounter_mapping, records)



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Optional, Union

import datetime


class Query:
    """
    Representation of the query portion of an ontology table, including table, column, type, operator and code
    """
    def __init__(self,
                 table: str,
                 key: str,
                 numeric_key: bool,
                 where_subj: str,
                 where_pred: str,
                 where_obj: Optional[Union[int, str, datetime.datetime]]) -> None:
        assert where_pred != 'like' or not numeric_key, "'like' queries not allowed on numeric keys"
        if where_obj is not None:
            self.table = table
            self.key = key
            self.numeric_key = numeric_key
            self.where_subj = where_subj
            self.where_pred = where_pred
            self.where_obj = where_obj
        else:
            self.table = self.key = self.where_subj = self.where_pred = self.where_obj = ""
            self.numeric_key = False

    def __str__(self):
        where_text = "{where_obj}" if self.numeric_key else "'{where_obj}'"
        return ("SELECT {key}\nFROM {table}\n"
                "WHERE {where_subj} {where_pred} " + where_text).format(**self.__dict__)


class EmptyQuery(Query):
    """
    No query - default values for all the elements
    """
    def __init__(self) -> None:
        super().__init__("", "", False, "", "", "")

    def __str__(self):
        return "NO QUERY"


class ConceptQuery(Query):
    """
    A query for the concept dimension.   The only (typical) variable is the object of the WHERE clause.
    """
    def __init__(self, where_obj: str) -> None:
        """
        Concept dimension query.
        Usually: "SELECT concept_cd FROM concept_dimension WHERE concept_path LIKE {where_obj}
        In postgresql, 'LIKE' w/o percent signs is equivalent to '=' so '=' is never needed
        :param where_obj: path to match. If none, will be set to fullpath in ontology
        """
        super().__init__("concept_dimension",
                         "concept_cd",
                         False,
                         "concept_path",
                         "=",
                         where_obj)


class ModifierQuery(Query):
    """
    Modifier dimension query.  The only (typical) variable is the object of the WHERE clause.
    """
    def __init__(self, where_obj: str) -> None:
        """
        Modifier dimension query.
        Usually: "SELECT modifier_cd FROM modifier_dimension WHERE modifier_path LIKE {where_obj}"
        In postgresql, 'LIKE' w/o percent signs is equivalent to '=' so '=' is never needed
        :param where_obj: path to match. If none, will be set to fullpath in ontology
        """
        super().__init__("modifier_dimension",
                         "modifier_cd",
                         False,
                         "modifier_path",
                         "like",
                         where_obj)


class PatientQuery(Query):
    """ Patient Dimension Query """
    def __init__(self,
                 where_subj: str,
                 where_pred: str,
                 where_obj: Union[int, str, datetime.datetime],
                 numeric_obj: bool = True):
        """
        Patient dimension query.
        Usually: "SELECT patient_num FROM patient_dimension WHERE {where_subj} {where_pred} {where_obj}"
        :param where_subj: patient_dimenension column name
        :param where_pred: predicate (aka. operator)
        :param where_obj:  target
        :param numeric_obj: True means target is numeric (or datetime).  False means string w/ quotes.
        """
        super().__init__("patient_dimension",
                         "patient_num",
                         numeric_obj,
                         where_subj,
                         where_pred,
                         where_obj)


class VisitQuery(Query):
    def __init__(self,
                 where_subj: str,
                 where_pred: str,
                 where_obj: Union[int, str, datetime.datetime],
                 numeric_obj: bool = True):
        """
        Visit dimension query.
        Usually: "SELECT encounter_num FROM visit_dimension WHERE {where_subj} {where_pred} {where_obj}"
        :param where_subj: visit_dimension column name
        :param where_pred: predicate (aka. operator)
        :param where_obj:  target
        :param numeric_obj: True means target is numeric (or datetime).  False means string w/ quotes.
        """
        super().__init__("visit_dimension",
                         "encounter_num",
                         numeric_obj,
                         where_subj,
                         where_pred,
                         where_obj)


class ProviderQuery(Query):
    def __init__(self, where_obj) -> None:
        """
        Provider dimension query.
        Usually: "SELECT provider_id FROM provider_dimension WHERE provider_path LIKE {where_obj}"
        In postgresql, 'LIKE' w/o percent signs is equivalent to '=' so '=' is never needed
        :param where_obj: path to match
        """
        super().__init__("provider_dimension",
                         "provider_id",
                         False,
                         'provider_id',
                         'like',
                         where_obj)



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

# TODO: The mapping portion of this function should be loaded from the i2b2 table mapping table
from typing import List

DEFAULT_ONTOLOGY_TABLE = "custom_meta"      # Default metadata ontology table


class _I2B2Tables:
    _funcs = {"phys_name", "all_tables"}

    def __init__(self) -> None:
        self.concept_dimension = None
        self.modifier_dimension = None
        self.table_access = None
        self.observation_fact = None
        self.ontology_table = DEFAULT_ONTOLOGY_TABLE
        self.patient_dimension = None
        self.patient_mapping = None
        self.visit_dimension = None
        self.provider_dimension = None
        self.encounter_mapping = None

    def __getattribute__(self, item):
        """ Return the logical name of a table  """
        if item.startswith("_") or item not in self.__dict__:
            return super().__getattribute__(item)
        return self.__dict__[item] if item in _I2B2Tables._funcs else item

    def __setattr__(self, key, value):
        super().__setattr__(key, value)

    def _clear(self):
        for k in self.all_tables():
            setattr(self, k, None if k != 'ontology_table' else DEFAULT_ONTOLOGY_TABLE)

    def phys_name(self, item: str) -> str:
        """Return the physical (mapped) name of item.

        :param item: logical table name
        :return: physical name of table
        """
        v = self.__dict__[item]
        return v if v is not None else item

    def all_tables(self) -> List[str]:
        """
        List of all known tables
        :return:
        """
        return sorted([k for k in self.__dict__.keys()
                       if k not in _I2B2Tables._funcs and not k.startswith("_")])


i2b2tablenames = _I2B2Tables()



# Copyright (c) 2018, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
import re
from argparse import Namespace
import shlex
from typing import List, Optional
import sys
import os

from i2fhirb2 import __version__
from i2fhirb2.common_cli_parameters import add_common_parameters
from i2fhirb2.sqlsupport.dbconnection import add_connection_args
from i2fhirb2.file_aware_parser import FileAwareParser

meta_parameters = ["file", "show", "create", "createorrep"]


def print_conf_file(opts: Namespace) -> Optional[str]:
    """ Print the config file contents

    :param opts: Options carrying file name
    :return: None if success otherwise error message
    """
    with open(opts.file) as f:
        args = shlex.split(f.read())
    parser = create_parser(use_defaults=False)
    opts = parser.parse_args(args)
    for k, v in opts.__dict__.items():
        if v is not None:
            print(f"{k}:\t\t{v}")
    return None


def create_conf_file(opts: Namespace) -> Optional[str]:
    """ Create the configuration file

    :param opts: Creation options
    :return: Error message if problems, otherwise None
    """
    if not opts.createorrep:
        if os.path.exists(opts.file):
            return f"{opts.file} already exists!"
    with open(opts.file, 'w') as f:
        for k, v in opts.__dict__.items():
            if k not in meta_parameters and v is not None:
                f.write(f"--{k} {esc(v)}\n")
    return None


def esc(arg: str) -> str:
    """ Escape arg in a way that the shlex (argv) processor keeps it whole

    :param arg: argument to escape
    :return: What to write in the file
    """
    arg = re.sub(r'\\', r'\\\\', str(arg))
    return f'"{arg}"' if re.search(r'\s', arg) else arg


def add_meta_args(parser: FileAwareParser) -> None:
    """ dd the control (meta) arguments to the parser

    :param parser: Parser to add arguments to
    """
    parser.add_argument("-f", "--file", help="Configuration file", default="db_conf")
    parser.add_argument("-s", "--show", help="Display current configuration", action="store_true")
    parser.add_argument("-c", "--create", help="Create new configuration file if not exists", action="store_true")
    parser.add_argument("-c!", "--createorrep", help="Create or replace configuration file ", action="store_true")


def create_parser(use_defaults: bool = True) -> FileAwareParser:
    """
    Create a command line argument parser

    :param use_defaults: Add defaults
    :return: parser
    """
    parser = FileAwareParser(description="Create configuration file for i2FHIRb2 software", prog="conf_file",
                             use_defaults=use_defaults)
    return add_connection_args(add_common_parameters(parser))


def conf_file(argv: List[str]) -> bool:
    """
    Create and/or print a configuration file

    :param argv: Parameter list -- see create_parser for details
    :return: True if success - otherwise an exception is thrown
    """
    parser = create_parser()
    add_meta_args(parser)
    opts = parser.parse_args(argv)

    # Note that "help" and "version" exit immediately
    if not (opts.create or opts.createorrep or opts.show):
        parser.print_usage()
        return False

    rval = create_conf_file(opts) if opts.create or opts.createorrep else None

    if not rval and opts.show:
        rval = print_conf_file(opts)

    if rval:
        parser.error(rval)

    return True


if __name__ == "__main__":
    conf_file(sys.argv[1:])



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Optional

from rdflib import URIRef, Graph
from rdflib.namespace import DC

from i2fhirb2.fhir.fhirspecific import concept_path, concept_code, modifier_code, modifier_name, \
    concept_name
from i2fhirb2.i2b2model.metadata.dimensionmetadata import metadata_xml
from i2fhirb2.i2b2model.metadata.i2b2ontologyquery import Query, ModifierQuery, ConceptQuery, EmptyQuery
from i2fhirb2.i2b2model.metadata.i2b2ontologyvisualattributes import VisualAttributes
from i2fhirb2.i2b2model.shared.i2b2core import I2B2Core
from i2fhirb2.sqlsupport.dynobject import DynElements, DynObject


class OntologyEntry(I2B2Core):
    _t = DynElements(I2B2Core)
    graph = None                # type: Graph
    ontology_name = "FHIR"

    def __init__(self,
                 c_full_name: str,
                 query: Query,
                 visualattributes: VisualAttributes = None,
                 c_basecode: Optional[str]=None,
                 primitive_type: Optional[URIRef]=None,
                 **kwargs):
        """
        Initialize an ontology entry.
        :param c_full_name: Full name of entry (e.g. '\\FHIR\\class\\subclass\\...\\item\\')
        :param query: Dimension table query for item
        :param visualattributes: VisualAttributes for item
        :param c_basecode: "uri" for item
        :param primitive_type: type used to construct c_metadataxml
        :param kwargs: Additional arguments for i2b2_core
        """
        super().__init__(**kwargs)
        assert(c_full_name.endswith('\\'))
        self._c_fullname = c_full_name
        self._query = query
        self._visualattributes = visualattributes if visualattributes else VisualAttributes()
        self._c_basecode = c_basecode
        self._modifier_exclusion = False
        self._primitive_type = primitive_type

    @DynObject.entry(_t)
    def c_hlevel(self) -> int:
        return self.c_fullname[:-1].count('\\') - 1

    @DynObject.entry(_t)
    def c_fullname(self) -> str:
        return self._c_fullname

    @DynObject.entry(_t)
    def c_name(self) -> str:
        return self.c_fullname[:-1].rsplit('\\', 1)[1]

    @DynObject.entry(_t)
    def c_synonym_cd(self) -> str:
        """ Two or more synonyms of each other will have the same c_basecode """
        return "N"

    @DynObject.entry(_t)
    def c_visualattributes(self) -> str:
        return str(self._visualattributes)

    @DynObject.entry(_t)
    def c_totalnum(self) -> Optional[int]:
        return None

    @DynObject.entry(_t)
    def c_basecode(self) -> Optional[str]:
        return self._c_basecode

    @DynObject.entry(_t)
    def c_metadataxml(self) -> Optional[str]:
        return metadata_xml(self._primitive_type, self.c_basecode, self.c_name, self.update_date) \
            if self._primitive_type else None

    @DynObject.entry(_t)
    def c_facttablecolumn(self) -> str:
        return self._query.key

    @DynObject.entry(_t)
    def c_tablename(self) -> str:
        return self._query.table

    @DynObject.entry(_t)
    def c_columnname(self) -> str:
        return self._query.where_subj

    @DynObject.entry(_t)
    def c_columndatatype(self) -> str:
        return 'N' if self._query.numeric_key else 'T'

    @DynObject.entry(_t)
    def c_operator(self) -> str:
        return self._query.where_pred

    @DynObject.entry(_t)
    def c_dimcode(self) -> str:
        return self._query.where_obj

    @DynObject.entry(_t)
    def c_comment(self) -> Optional[str]:
        return None

    @DynObject.entry(_t)
    def c_tooltip(self) -> Optional[str]:
        return None

    @DynObject.entry(_t)
    def m_applied_path(self) -> str:
        return '@'

    DynObject._after_root(_t)

    @DynObject.entry(_t)
    def valuetype_cd(self) -> Optional[str]:
        return None

    @DynObject.entry(_t)
    def m_exclusion_cd(self) -> Optional[str]:
        return 'X' if self._modifier_exclusion else None

    @DynObject.entry(_t)
    def c_path(self) -> Optional[str]:
        # return self.basename[:-1].rsplit('\\', 1)[0]
        return None

    @DynObject.entry(_t)
    def c_symbol(self) -> Optional[str]:
        # return self.basename[:-1].rsplit('\\', 1)[1]
        return None

    def __lt__(self, other):
        return self.c_fullname + self.m_applied_path < other.c_fullname + self.m_applied_path

    def __eq__(self, other):
        return self.c_fullname + self.m_applied_path == other.c_fullname + self.m_applied_path


class OntologyRoot(OntologyEntry):
    _t = DynElements(OntologyEntry)

    def __init__(self, base: str, **kwargs) -> None:
        """
        Initialize an ontology header.
        :param kwargs: Additional arguments for i2b2_core
        """
        if base.startswith('\\'):
            base = base[1:-1]
        path = '\\' + base + '\\'
        super().__init__(path, EmptyQuery(), VisualAttributes("CA"), sourcesystem_cd=base, **kwargs)
        self._base = base

    @DynObject.entry(_t)
    def c_hlevel(self) -> int:
        return 0

    @DynObject.entry(_t)
    def c_name(self) -> str:
        return self._base

    @DynObject.entry(_t)
    def c_basecode(self) -> Optional[str]:
        return self._base + ':'


class ModifierOntologyEntry(OntologyEntry):
    _t = DynElements(OntologyEntry)

    def __init__(self,
                 depth: int,
                 subject: URIRef,
                 mod: URIRef,
                 full_path: str,
                 applied_path: str,
                 is_leaf: bool,
                 mod_pred: URIRef,
                 mod_type: URIRef) -> None:
        """
        Construct a concept entry in the ontology space
        :param depth: Relative one-based depth of entry
        :param subject: URI of concept being modified
        :param mod: URI of modifier itself
        :param full_path: modifier path
        :param applied_path: path that modifier applies to
        :param is_leaf: true if there are no additional children
        :param mod_pred: real modifier predicate (for dimcode)
        :param mod_type: Actual type of path
        """
        assert(depth > 0)

        # TODO: Fix the hard coded reference below
        query = ModifierQuery('\\FHIR\\' + concept_path(mod_pred)) if is_leaf else EmptyQuery()

        visattr = VisualAttributes()
        visattr.leaf = is_leaf
        visattr.concept = False
        visattr.draggable = True
        visattr.editable = False

        super().__init__(full_path, query, visattr, modifier_code(mod), mod_type)

        self._subject = subject
        self._mod = mod
        self._depth = depth
        self._m_applied_path = applied_path
        self._mod_pred = mod_pred

    # Levels in ontology modifier references start at 1
    @DynObject.entry(_t)
    def c_hlevel(self) -> int:
        return self._depth

    @DynObject.entry(_t)
    def c_name(self) -> str:
        return modifier_name(self.graph, self._mod)

    @DynObject.entry(_t)
    def c_comment(self) -> Optional[str]:
        rval = self.graph.comment(self._mod)
        if not rval and self._primitive_type:
            rval = self.graph.comment(self._primitive_type)
        return str(rval) if rval else None

    @DynObject.entry(_t)
    def c_tooltip(self) -> Optional[str]:
        rval = self.graph.value(self._mod, DC.title, None, self.graph.comment(self._mod))
        if not rval and self._primitive_type:
            rval = self.graph.value(self._primitive_type, DC.title, None, self.graph.comment(self._primitive_type))
        return str(rval) if rval else None

    @DynObject.entry(_t)
    def m_applied_path(self) -> str:
        return self._m_applied_path


class ConceptOntologyEntry(OntologyEntry):
    _t = DynElements(OntologyEntry)

    def __init__(self,
                 subject: URIRef,
                 navigational_path: str,
                 ontological_path: str,
                 is_leaf: bool,
                 is_draggable: bool=True,
                 primitive_type: Optional[URIRef]=None):
        """
        Construct a concept entry in the ontology space
        :param subject: URI of concept
        :param navigational_path: path to concept for navigational purposes.
          Example: \\FHIR\\administrative\\individual\\Patient
        :param ontological_path: concept_dimension path to concept.  Example: \\FHIR\\Patient
        :param is_leaf: true if there are no additional children.
        :param is_draggable: If not a leaf, whether this is an i2b2 container (False) or folder (True)
        :param primitive_type: Type used to construct c_metadataxml
        """
        self._subject = subject

        visattr = VisualAttributes()
        visattr.leaf = is_leaf
        visattr.concept = True
        visattr.draggable = is_draggable
        visattr.editable = False

        full_path = navigational_path
        self._m_applied_path = '@'

        query = ConceptQuery(ontological_path) if is_draggable else EmptyQuery()
        super().__init__(full_path, query, visattr, concept_code(subject), primitive_type)

    # Level in ontology concept references are relative to base path
    @DynObject.entry(_t)
    def c_hlevel(self) -> int:
        return self.c_fullname[:-1].count('\\') - 1

    @DynObject.entry(_t)
    def c_name(self) -> str:
        return concept_name(self.graph, self._subject)

    @DynObject.entry(_t)
    def c_comment(self) -> Optional[str]:
        rval = self.graph.comment(self._subject)
        return str(rval) if rval else None

    @DynObject.entry(_t)
    def c_tooltip(self) -> Optional[str]:
        rval = self.graph.value(self._subject, DC.title, None, self.graph.comment(self._subject))
        return str(rval) if rval else None



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
import sys
from typing import Dict, List, Union, Tuple, NamedTuple

from fhirtordf.rdfsupport.namespaces import FHIR
from rdflib import Graph, URIRef, RDFS, RDF, OWL, BNode
from rdflib.collection import Collection

from i2fhirb2.fhir.fhirspecific import is_primitive, skip_fhir_predicates

# TODO: The FMV needs to be updated to add a type element -- we use this enumeration as a temporary solution
fhir_complex_types = {FHIR.Address, FHIR.Age, FHIR.Annotation, FHIR.Attachment,
                      FHIR.CodeableConcept, FHIR.Coding, FHIR.ContactDetail, FHIR.ContactPoint,
                      FHIR.Count, FHIR.Distance,  FHIR.Duration, FHIR.HumanName,
                      FHIR.Identifier, FHIR.Money, FHIR.Period, FHIR.Quantity, FHIR.Range, FHIR.Ratio,
                      FHIR.Signature, FHIR.SimpleQuantity, FHIR.Timing}


class FMVGraphEdge:
    def __init__(self, g: Graph, restriction_node: Union[URIRef, BNode]) -> None:
        """
        An edge (predicate) in an FMV graph.  Determined by parsing either an existential (OWL:someValuesFrom)
         or universal (OWL:allValuesFrom) element in an OWL:Restriction node.

        Elements:
        * predicate -- URIRef of restriction property (example: FHIR.Observation.status)
        * multiple_entries -- True if element can occur more than one time
        * type_node -- target of edge

        :param g: Graph containing FMV
        :param restriction_node: Restriction entry
        """
        self.predicate, predicate_type = FMVGraphNode.predicate_and_type_for(g, restriction_node)
        self.multiple_entries = int(g.value(restriction_node, OWL.maxCardinality,
                                            default=int(g.value(restriction_node, OWL.cardinality,
                                                                default=sys.maxsize)))) > 1
        if predicate_type not in FMVGraphNode.known_nodes:
            self.type_node = FMVGraphNode(g, predicate_type)
        else:
            self.type_node = FMVGraphNode.known_nodes[predicate_type]

    def as_indented_str(self, indent: int) -> str:
        rval = 4 * indent * ' '
        if indent > 10:
            return rval + '   ...'
        else:
            return rval + FMVGraphNode.fns(self.predicate) + " " + \
                self.type_node.as_indented_str(indent, self.multiple_entries)

    def __lt__(self, other: "FMVGraphEdge") -> bool:
        return str(self.predicate) < str(other.predicate)

    def __eq__(self, other: "FMVGraphEdge") -> bool:
        return str(self.predicate) == str(other.predicate)

    def __hash__(self):
        return hash((self.predicate, self.type_node))


class FMVGraphNode:
    known_nodes = {}        # type: Dict[URIRef, 'FMVGraphNode']

    def __init__(self, g: Graph, node: URIRef) -> None:
        """
        Construct a directed graph rooted in node

        :param g: FHIR Metadata Vocabulary + w5
        :param node: URI of node
        """
        self.node: URIRef = node                            # URI of the node itself
        self.edges: List[FMVGraphEdge] = []                 # Outgoing edges
        self.parents: List["FMVGraphNode"] = []                     # Parent types
        FMVGraphNode.known_nodes[node]: Dict[URIRef, FMVGraphNode] = self  # Prevent recursive looping
        self.is_primitive: bool = is_primitive(g, node)

        if not self.is_primitive:
            self.is_complex_type = node in fhir_complex_types
            for sc in g.objects(node, RDFS.subClassOf):
                unionof_value = g.value(sc, OWL.unionOf)
                if unionof_value:
                    for union_sc in Collection(g, unionof_value):
                        predicate, _ = self.predicate_and_type_for(g, union_sc)
                        if predicate not in skip_fhir_predicates:
                            self.edges.append(FMVGraphEdge(g, union_sc))
                else:
                    for sc_type in g.objects(sc, RDF.type):
                        if sc_type == OWL.Restriction:
                            predicate, _ = self.predicate_and_type_for(g, sc)
                            if predicate not in skip_fhir_predicates:
                                self.edges.append(FMVGraphEdge(g, sc))
                        elif sc_type != OWL.Class:
                            self.edges += self.known_nodes[sc_type].edges \
                                if sc_type in FMVGraphNode.known_nodes else FMVGraphNode(g, sc_type).edges
                        else:
                            self.parents.append(self.known_nodes[sc]
                                                if sc in FMVGraphNode.known_nodes else FMVGraphNode(g, sc))
        else:
            self.is_complex_type = False

    @staticmethod
    def predicate_and_type_for(g: Graph, restriction_node: Union[URIRef, BNode]) -> Tuple[URIRef, URIRef]:
        predicate = g.value(restriction_node, OWL.onProperty)
        predicate_type = g.value(restriction_node, OWL.allValuesFrom)
        if predicate_type is None:
            predicate_type = g.value(restriction_node, OWL.someValuesFrom)
        return predicate, predicate_type

    def as_indented_str(self, indent: int=0, multiple_entries: bool = False) -> str:
        """
        Pretty-print node as an indented string
        :param indent: depth at this point
        :param multiple_entries: true means the cardinality > 1
        :return: indented string
        """
        modifiers = ("M" if multiple_entries else "") + ("P" if self.is_primitive else "") + \
                    ("C" if self.is_complex_type else "")
        return "({})".format(self.fns(self.node)) + ("({})".format(modifiers) if modifiers else "") + \
               ((':\n' + '\n'.join([e.as_indented_str(indent+1) for e in sorted(self.edges)])) if self.edges else "")

    @staticmethod
    def fns(uri: URIRef) -> str:
        """
        Change URI into fhir:name if appropriate

        :param uri: URI to convert
        :return: ns/name form if FHIR URI else turtle form
        """
        str_uri = str(uri)
        return str_uri.replace(str(FHIR), 'fhir:') if str(FHIR) in str_uri else "<{}>".format(str_uri)

    def __str__(self) -> str:
        return self.as_indented_str()


class FMVGraphNodeWithMultiplicity(NamedTuple):
    graph_node: "FMVGraphNode"
    is_multiple: bool



# Copyright (c) 2018, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Union, ChainMap, Optional

from fhirtordf.rdfsupport.namespaces import AnonNS, namespaces as fhirtordf_namespaces, FHIR, SCT
from fhirtordf.rdfsupport.numericnamespace import NumericNamespace
from rdflib import Namespace, URIRef


HGNC = Namespace("http://www.genenames.org/")
RXNORM = NumericNamespace("http://www.nlm.nih.gov/research/umls/rxnorm/")

# Note that this is a different namespace than the one that is included in the type arc
SNOMED = Namespace("http://snomed.info/sct/")


namespaces = ChainMap(fhirtordf_namespaces,
                      {"hgnc": str(HGNC),
                       "rxnorm": str(RXNORM)
                       })


def fhir_namespace_for(uri: Union[URIRef, Namespace, str]) -> Optional[str]:
    """
    Reverse namespace lookup.  Note that returned namespace may not be unique

    :param uri: namespace URI
    :return: namespace (lower case form)
    """
    uri = str(uri)
    if not uri.endswith(('#', '/')):
        uri += '/'
    if uri not in namespaces.values():
        if uri.startswith(str(FHIR)):
            ns = uri[len(str(FHIR)):-1]
            namespaces[ns] = uri
        elif uri == str(SNOMED):
            return fhir_namespace_for(SCT)
        else:
            return None
    return [k for k, v in namespaces.items() if uri == v][0]




# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
import sys
from argparse import Namespace
from typing import List, Tuple, Optional

from sqlalchemy import Table, delete
from sqlalchemy.engine import Connection
from sqlalchemy.orm import sessionmaker, Session

from i2fhirb2.common_cli_parameters import add_common_parameters
from i2fhirb2.file_aware_parser import FileAwareParser
from i2fhirb2.loaders.i2b2graphmap import I2B2GraphMap
from i2fhirb2.sqlsupport.dbconnection import add_connection_args, process_parsed_args, I2B2Tables


def create_parser() -> FileAwareParser:
    """
    Create a command line parser
    :return: parser
    """
    parser = FileAwareParser(description="Clear data from FHIR observation fact table", prog="removefacts",
                             use_defaults=False)
    add_connection_args(add_common_parameters(parser, multi_upload_ids=True), strong_config_file=False)
    parser.add_argument("--testlist", help="List leftover test suite entries", action="store_true")
    parser.add_argument("--removetestlist", help="Remove leftover test suite entries", action="store_true")
    return parser


def sourcesystem_test_query(c: Connection, t: Table) -> List[Tuple[Table, str]]:
    """ Generate a sourcesystem_cd test query for table t

    :param c: Connection
    :param t: Table
    :return: list of elements
    """
    q = c.query(t.c.sourcesystem_cd).filter(t.c.sourcesystem_cd.startswith('test_i2FHIRb2_')).distinct()
    return [(t, e[0]) for e in q.all()]


def list_test_artifacts(opts: Optional[Namespace], tables: Optional[I2B2Tables]=None) -> List[Tuple[Table, str]]:
    """  Return a list of tables and sourcesystem_ids from the test suite

    :param opts: Passed options.  Can be omitted if tables is present
    :param tables: I2B2 tables
    :return: List 
    """
    if tables is None:
        tables = I2B2Tables(opts)
    session = sessionmaker(bind=tables.crc_engine)()
    qr = sourcesystem_test_query(session, tables.patient_dimension)
    qr += sourcesystem_test_query(session, tables.patient_mapping)
    qr += sourcesystem_test_query(session, tables.visit_dimension)
    qr += sourcesystem_test_query(session, tables.encounter_mapping)
    qr += sourcesystem_test_query(session, tables.provider_dimension)
    qr += sourcesystem_test_query(session, tables.observation_fact)
    if qr:
        print('\n'.join(f"TABLE: {e[1]} \t: {e[0]}" for e in qr))
    return qr


def remove_test_artifacts(opts: Namespace):
    """ Remove any test artifacts (sourcesystem_id starts with 'test_i2FHIRb2_')
    
    :param opts: 
    :return: 
    """
    tables = I2B2Tables(opts)
    artifacts_list = list_test_artifacts(opts, tables)
    for table, ss_cd in artifacts_list:
        q = delete(table).where(table.c.sourcesystem_cd == ss_cd)
        ndel = opts.tables.crc_connection.execute(q).rowcount
        print(f"{ndel} rows removed from {table}")
    return True


def remove_facts(argv: List[str]) -> bool:
    """
    Convert a set of FHIR resources into their corresponding i2b2 counterparts.
    
    :param argv: Command line arguments.  See: create_parser for details
    :return:
    """
    parser = create_parser()
    local_opts = parser.parse_args(argv)                        # Pull everything from the actual command line
    if not (local_opts.uploadid or local_opts.sourcesystem or local_opts.testlist or local_opts.removetestlist):
        parser.error("Option must be one of: -ss, -u, --testlist, --removetestlist)")

    opts = parser.parse_args(parser.decode_file_args(argv))     # Include the options file
    if opts is None:
        return False
    opts.uploadid = local_opts.uploadid
    opts.sourcesystem = local_opts.sourcesystem

    process_parsed_args(opts, parser.error)           # Update CRC and Meta table connection information

    if opts.uploadid:
        for uploadid in opts.uploadid:
            print("---> Removing entries for id {}".format(uploadid))
            I2B2GraphMap.clear_i2b2_tables(I2B2Tables(opts), uploadid)
    if opts.sourcesystem:
        print("---> Removing entries for sourcesystem_cd {}".format(opts.sourcesystem))
        I2B2GraphMap.clear_i2b2_sourcesystems(I2B2Tables(opts), opts.sourcesystem)
    if opts.testlist:
        print("---> Listing orphan test elements in database")
        list_test_artifacts(opts)
    if opts.removetestlist:
        print("---> Removing orphan test elements in database")
        remove_test_artifacts(opts)
    return True


if __name__ == "__main__":
    remove_facts(sys.argv[1:])



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
import os
import sys
from argparse import Namespace
from datetime import datetime
from random import randint
from typing import List, Optional
from urllib.request import Request, urlopen
from i2fhirb2 import __version__

from fhirtordf.fhir.fhirmetavoc import FHIRMetaVoc
from fhirtordf.loaders.fhirjsonloader import fhir_json_to_rdf
from rdflib import Graph

from i2fhirb2.common_cli_parameters import add_common_parameters
from i2fhirb2.fhir.fhirencountermapping import FHIREncounterMapping
from i2fhirb2.fhir.fhirpatientmapping import FHIRPatientMapping
from i2fhirb2.i2b2model.data.i2b2observationfact import ObservationFact

from i2fhirb2.i2b2model.data.i2b2patientdimension import PatientDimension
from i2fhirb2.i2b2model.data.i2b2patientmapping import PatientMapping
from i2fhirb2.i2b2model.shared.i2b2core import I2B2CoreWithUploadId


from i2fhirb2.loaders.i2b2graphmap import I2B2GraphMap
from i2fhirb2.sqlsupport.dbconnection import add_connection_args, process_parsed_args, I2B2Tables
from i2fhirb2.file_aware_parser import FileAwareParser


# TODO: Add support for non-turtle RDF files
# TODO: Add continuation headers for RDF


def load_rdf_graph(opts: Namespace) -> Optional[Graph]:
    """
    Load the file(s) specified by opts into an RDF graph
    :param opts: User supplied option
    :return: Loaded graph or None if errors were encountered
    """
    g = Graph()
    fmv = FHIRMetaVoc(opts.metadatavoc + '/fhir.ttl')

    def read_rdf_uri(uri: str) -> str:
        req = Request(uri)
        req.add_header("Accept", "application/turtle, text/turtle;q=0.9")
        with urlopen(req) as response:
            return response.read().decode()

    def load_file(dirname: str, fname: str) -> None:
        filepath = fname if '://' in fname else os.path.join(dirname, fname)
        print("--> loading {}".format(filepath))
        if '://' in fname:
            if opts.filetype == 'rdf':
                rdf_str = read_rdf_uri(fname)
                if not rdf_str:
                    print("   Read Failed")
                g.parse(data=rdf_str, format="turtle")
            else:
                fhir_json_to_rdf(fname, opts.uribase, g, metavoc=fmv)
        else:
            if filepath.endswith('.json'):
                fhir_json_to_rdf(filepath, opts.uribase, g, metavoc=fmv)
            else:
                g.load(filepath, format="turtle")

    if opts.infile:
        [load_file(opts.indir if opts.indir else "", fn) for fn in opts.infile]
    else:
        for dirpath, _, filenames in os.walk(opts.indir):
            for filename in filenames:
                if (opts.filetype == 'json' and filename.endswith('.json')) or filename.endswith('.ttl'):
                    load_file(dirpath, filename)
    return g


def create_parser() -> FileAwareParser:
    """
    Create a command line argument parser
    :return: parser
    """
    parser = FileAwareParser(description="Load FHIR Resource Data into i2b2 CRC tables", prog="loadfacts")
    parser.add_argument("-l", "--load", help="Load SQL Tables", action="store_true")
    parser.add_file_argument("-i", "--infile",
                             metavar="Input files", help="URLs and/or name(s) of input file(s)", nargs='*')
    parser.add_file_argument("-id", "--indir", metavar="Input directory",
                             help="URI of server or directory of input files")
    parser.add_file_argument("-od", "--outdir", metavar="Output directory",
                             help="Output directory to store .tsv files.")
    parser.add_argument("-t", "--filetype",
                        help="Type of file to ask for / load - only applies for URL's and directories.",
                        choices=['json', 'rdf'], default='rdf')
    parser.add_argument("-rm", "--remove", help="Remove existing entries for the upload identifier and/or"
                        " clear target tsv files", action="store_true")
    parser.add_argument("--dupcheck", help="Check for duplicate records before add.", action="store_true")
    return add_common_parameters(parser)


def genargs(argv: List[str]) -> Optional[Namespace]:
    """
    Parse the input arguments and create the options list
    :param argv: input arguments
    :return: options if success or None of parameters aren't valid
    """
    parser = add_connection_args(create_parser())
    opts = parser.parse_args(parser.decode_file_args(argv))
    if opts.version:
        print("FHIR i2b2 CRC loader -- Version {}".format(__version__))
    elif not (opts.load or opts.outdir):
        parser.error("Either load option (-l) or output directory must be specified")
    if not (opts.infile or opts.indir or opts.version):
        parser.error("Either a list of input files or input directory must be supplied")
    if opts.remove and not opts.load:
        parser.error("Remove existing upload id only implemented for LOAD option")
    if opts.infile:
        for fn in opts.infile:
            if '://' not in fn and not (fn.endswith('.ttl') or fn.endswith(".json")):
                parser.error("Unrecognized file type: {}".format(fn))
    if opts.load or opts.outdir:
        if not opts.uploadid:
            # TODO: find a more rational way to do this
            # TODO: uploadid to description map
            opts.uploadid = randint(200000, 500000)
        opts.updatedate = datetime.now()
        if opts.indir and not opts.indir.endswith(os.sep):
            opts.indir = os.path.join(opts.indir, '')
        if opts.outdir and not opts.outdir.endswith(os.sep):
            opts.outdir = os.path.join(opts.outdir, '')
        if opts.load:
            process_parsed_args(opts, parser.error)
        I2B2CoreWithUploadId.sourcesystem_cd = opts.sourcesystem
        I2B2CoreWithUploadId.upload_id = opts.uploadid
        return opts
    return None


def print_rdf_summary(g: Graph()) -> None:
    """
    Summarize the number of resources and other information that was actually loaded
    :param g: Graph of loaded RDF
    """
    # TODO: Figure out how to count the number of resources
    # num_resources = len(list(g.subject_objects(FHIR.resourceType)))
    num_triples = len(g)
    # print("Loaded ??? resources creating {} triples (Unable to determine how many...)".format(num_triples))
    print("{} triples".format(num_triples))


def load_graph_map(opts: Namespace) -> Optional[I2B2GraphMap]:
    """
    Transform the input URI(s) and/or file(s) into an I2B2GraphMap
    :param opts: input options
    :return: I2B2GraphMap if success otherwise None
    """
    opts.tables = I2B2Tables(opts) if opts.load else None
    print("upload_id: {}".format(opts.uploadid))
    if opts.tables:
        print("  Starting encounter number: {}"
              .format(FHIREncounterMapping.refresh_encounter_number_generator(opts.tables,
                                                                              opts.uploadid if opts.remove else None)))
        print("  Starting patient number: {}"
              .format(FHIRPatientMapping.refresh_patient_number_generator(opts.tables,
                                                                          opts.uploadid if opts.remove else None)))
    g = load_rdf_graph(opts)
    if g:
        update_dt = datetime.now()
        update_dt_coarse = datetime(update_dt.year, update_dt.month, update_dt.day, update_dt.hour, update_dt.minute)
        ObservationFact.update_date = update_dt_coarse
        PatientDimension.update_date = update_dt_coarse
        PatientMapping.update_date = update_dt_coarse
        print_rdf_summary(g)
        return I2B2GraphMap(g, opts)
    return None


def load_facts(argv: List[str]) -> bool:
    """
    Convert a set of FHIR resources into their corresponding i2b2 counterparts.
    :param argv: Command line arguments.  See: create_parser for details
    :return:
    """
    opts = genargs(argv)
    if not opts:
        return False
    else:
        i2b2_map = load_graph_map(opts)
        if not i2b2_map:
            return False
        else:
            print(i2b2_map.summary())
            if opts.outdir:
                i2b2_map.generate_tsv_files()
            if opts.load:
                i2b2_map.load_i2b2_tables(opts.dupcheck)
            return True


if __name__ == "__main__":
    load_facts(sys.argv[1:])



# Copyright (c) 2018, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
import argparse
import shlex
from typing import List

import os


class FileAwareParser(argparse.ArgumentParser):
    """ Argument parser that tracks which arguments are file names for relative file relocation. """
    def __init__(self, use_defaults: bool=True, *args, **kwargs) -> None:
        """ Constructor

        :param use_defaults: If False, omit defaults values.  Used for parsing configuration files
        :param args: ArgumentParser args -- passed on
        :param kwargs: ArgumentParser args -- passed on
        """
        self.use_defaults = use_defaults
        super().__init__(*args, **kwargs)
        self.file_args = []

    def add_file_argument(self, *args, **kwargs):
        """ Add an argument that represents the location of a file

        :param args:
        :param kwargs:
        :return:
        """
        rval = self.add_argument(*args, **kwargs)
        self.file_args.append(rval)
        return rval

    def add_argument(self, *args, **kwargs):
        """ Add an argument incorporating the default value into the help string

        :param args:
        :param kwargs:
        :return:
        """
        defhelp = kwargs.pop("help", None)
        defaults = kwargs.pop("default", None)
        default = defaults if self.use_defaults else None
        if not defhelp or default is None or kwargs.get('action') == 'help':
            return super().add_argument(*args, help=defhelp, default=default, **kwargs)
        else:
            return super().add_argument(*args, help=defhelp + " (default: {})".format(default),
                                        default=default, **kwargs)

    def decode_file_args(self, argv: List[str]) -> List[str]:
        """
        Preprocess a configuration file.  The location of the configuration file is stored in the parser so that the
        FileOrURI action can add relative locations.
        :param argv: raw options list
        :return: options list with '--conf' references replaced with file contents
        """
        for i in range(0, len(argv) - 1):
            # TODO: take prefix into account
            if argv[i] == '--conf':
                del argv[i]
                conf_file = argv[i]
                del (argv[i])
                with open(conf_file) as config_file:
                    conf_args = shlex.split(config_file.read())
                    # We take advantage of a poential bug in the parser where you can say "foo -u 1 -u 2" and get
                    # 2 as a result
                    argv = self.fix_rel_paths(conf_args, conf_file) + argv
                return self.decode_file_args(argv)
        return argv

    def fix_rel_paths(self, conf_args: List[str], conf_file: str) -> List[str]:
        base_path = os.path.abspath(os.path.split(conf_file)[0])
        rval = []
        is_file_arg = False
        for conf_arg in conf_args:
            if any(conf_arg.startswith(prefix) for prefix in self.prefix_chars):
                rval.append(conf_arg)
                is_file_arg = any(conf_arg in file_action.option_strings for file_action in self.file_args)
                # TODO: recursive config files
            elif is_file_arg and ('://' not in conf_arg and not os.path.isabs(conf_arg)):
                rval.append(os.path.abspath(os.path.join(base_path, conf_arg)))
            else:
                rval.append(conf_arg)
        return rval



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Dict, Tuple, Optional

from fhirtordf.rdfsupport.uriutils import parse_fhir_resource_uri
from rdflib import URIRef
from rdflib.namespace import split_uri
from sqlalchemy import func, or_
from sqlalchemy.orm import sessionmaker

from i2fhirb2.common_cli_parameters import DEFAULT_PROJECT_ID, DEFAULT_ENCOUNTER_NUMBER_START, IDE_SOURCE_HIVE
from i2fhirb2.i2b2model.data.i2b2encountermapping import EncounterMapping, EncounterIDEStatus
from i2fhirb2.sqlsupport.dbconnection import I2B2Tables


class EncounterNumberGenerator:
    """
    i2b2 encounter number generator.
    """
    def __init__(self, next_number: int) -> None:
        """
        Create the number generator
        :param next_number: First encounter number to assign
        """
        self._next_number = next_number

    def new_number(self) -> int:
        """
        Return a new encounter number
        :return:
        """
        rval = self._next_number
        self._next_number += 1
        return rval

    def refresh(self, tables: I2B2Tables, ignore_upload_id: Optional[int]) -> int:
        """
        Determine the greatest encounter number that is currently in use and set the next number to one greater
        :param tables: i2b2 SQL tables
        :param ignore_upload_id: If present, encounters with this upload id are ignored.  This is used in cases when
        an upload is being replaced.
        :return: next number
        """
        session = sessionmaker(bind=tables.crc_engine)()
        q = func.max(tables.visit_dimension.c.encounter_num)
        if ignore_upload_id is not None:
            q = q.filter(or_(tables.visit_dimension.c.upload_id.is_(None),
                         tables.visit_dimension.c.upload_id != ignore_upload_id))
        qr = session.query(q).all()
        self._next_number = qr[0][0] + 1
        session.close()
        return self._next_number


class FHIREncounterMapping:
    project_id = DEFAULT_PROJECT_ID         # Project identifier
    identity_source_id = IDE_SOURCE_HIVE    # source_id for identity mapping
    number_generator = EncounterNumberGenerator(DEFAULT_ENCOUNTER_NUMBER_START)
    number_map = dict()                     # type: Dict[Tuple[str, str, str, str, str], int]

    @classmethod
    def _clear(cls) -> None:
        """
        Reset the mapping table to its default. (Primarily used for testing)
        """
        cls.project_id = DEFAULT_PROJECT_ID
        cls.identity_source_id = IDE_SOURCE_HIVE
        cls.number_generator = EncounterNumberGenerator(DEFAULT_ENCOUNTER_NUMBER_START)
        cls.number_map.clear()

    @classmethod
    def refresh_encounter_number_generator(cls, tables: I2B2Tables, ignore_upload_id: Optional[int]) -> int:
        """
        Reset the generator
        :param tables: i2b2 tables
        :param ignore_upload_id: if present, do not look at the encounter numbers from this upload
        :return: starting encounter number
        """
        return cls.number_generator.refresh(tables, ignore_upload_id)

    def __init__(self, encounterURI: URIRef, patient_id: str, patient_ide_source: str) -> None:
        """
        Create a new encounter mapping entry
        :param encounterURI: URI of the encounter
        :param patient_id: Associated patient identifier
        :param patient_ide_source: Associated patient identifier source
        """
        self.encounter_mapping_entries = []
        parsed_resource = parse_fhir_resource_uri(encounterURI)
        resource_namespace = str(parsed_resource.namespace)
        resource_ide = split_uri(parsed_resource.resource_type)[1] + '/' + parsed_resource.resource
        key = (resource_ide, resource_namespace, self.project_id, patient_id, patient_ide_source)
        if key in self.number_map:
            self.encounter_num = self.number_map[key]
        else:
            self.encounter_num = self.number_generator.new_number()
            pm = EncounterMapping(resource_ide, resource_namespace, self.project_id,
                                  self.encounter_num, patient_id, patient_ide_source, EncounterIDEStatus.active)
            self.number_map[key] = self.encounter_num
            self.encounter_mapping_entries.append(pm)

        identity_id = str(self.encounter_num)
        ikey = (identity_id, self.identity_source_id, self.project_id)
        if ikey not in self.number_map:
            ipm = EncounterMapping(identity_id,
                                   self.identity_source_id,
                                   self.project_id,
                                   self.encounter_num,
                                   patient_id,
                                   patient_ide_source,
                                   EncounterIDEStatus.active)
            self.encounter_mapping_entries.append(ipm)



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Sized

import sys


class ListChunker:
    """
    Wrapper to iterate over potentially large lists in chunks.  We use this to commit large volumes of data in
    batches.
    """
    def __init__(self, lst: Sized, interval: int, print_progress: bool = True) -> None:
        """
        Iterator that returns elements of lst in chunks of size interval or less
        :param lst: List to be returned in batches
        :param interval: chunk size
        :param print_progress: True means print a '.' on stdout per chunk
        """
        self.input = lst
        self.interval = interval
        self._pos = 0
        self._dotted = False
        self._print_progress = print_progress

    def __iter__(self):
        return self

    def __next__(self) -> Sized:
        if self._pos < len(self.input):
            start = self._pos
            self._pos = self._pos + self.interval
            if self._pos < len(self.input) and self._print_progress:
                print('.', end='')
                sys.stdout.flush()
                self._dotted = True
            return self.input[start:self._pos]
        else:
            if self._dotted:
                print()
            raise StopIteration
